#ifndef MULTILEVEL_PATCH_PRECONDITIONER_HH
#define MULTILEVEL_PATCH_PRECONDITIONER_HH

#include <string>

#include <dune/common/timer.hh>
#include <dune/common/fvector.hh>
#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>

#include <dune/solvers/norms/energynorm.hh>
#include <dune/solvers/solvers/loopsolver.hh>
#include <dune/solvers/solvers/umfpacksolver.hh>
#include <dune/solvers/iterationsteps/blockgssteps.hh>
//#include <dune/solvers/iterationsteps/blockgsstep.hh>
#include <dune/solvers/iterationsteps/cgstep.hh>
#include <dune/solvers/iterationsteps/lineariterationstep.hh>

#include "nbodycontacttransfer.hh"
#include "levelpatchpreconditioner.hh"

#include "../../utils/debugutils.hh"

template <class ContactNetwork, class MatrixType, class VectorType>
class MultilevelPatchPreconditioner : public LinearIterationStep<MatrixType, VectorType> {
public:
    //struct Mode { enum ModeType {additive, multiplicative}; };

private:
    using Base = LinearIterationStep<MatrixType, VectorType>;

    //using PatchSolver = typename Dune::Solvers::LoopSolver<Vector, BitVector>;
    using PatchSolver = Dune::Solvers::UMFPackSolver<MatrixType, VectorType>;

    using LevelContactNetwork = typename ContactNetwork::LevelContactNetwork;
    using LevelPatchPreconditioner = LevelPatchPreconditioner<LevelContactNetwork, PatchSolver, MatrixType, VectorType>;
    using BitVector = typename LinearIterationStep<MatrixType, VectorType>::BitVector;

    using MGTransfer = NBodyContactTransfer<ContactNetwork, VectorType>;

    static const int dim = ContactNetwork::dim;

    const ContactNetwork& contactNetwork_;
    const Dune::BitSetVector<1>& activeLevels_;

    const MPPMode mode_;

    std::vector<std::shared_ptr<LevelPatchPreconditioner>> levelPatchPreconditioners_;

    std::vector<MatrixType> levelMat_;
    std::vector<VectorType> levelX_;
    std::vector<VectorType> levelRhs_;

    std::vector<BitVector> ignoreHierarchy_;

    std::vector<std::shared_ptr<EnergyNorm<MatrixType, VectorType>>> levelErrorNorms_;
    std::vector<std::shared_ptr<LinearIterationStep<MatrixType, VectorType>>> levelItSteps_;
    std::vector<std::shared_ptr<PatchSolver>> levelSolvers_;

    std::vector<BitVector> recompute_;
    //std::vector<BitVector> truncation_;
    std::vector<std::shared_ptr<MGTransfer>> mgTransfer_; //TODO

public:
    MultilevelPatchPreconditioner(const Dune::ParameterTree& parset,
                                  const ContactNetwork& contactNetwork,
                                  const Dune::BitSetVector<1>& activeLevels) :
          contactNetwork_(contactNetwork),
          activeLevels_(activeLevels),
          mode_(parset.get<MPPMode>("mode")){

        assert(activeLevels.size() == contactNetwork_.nLevels());

        // init level patch preconditioners and multigrid transfer
        levelPatchPreconditioners_.resize(1);
        levelPatchPreconditioners_[0] = nullptr; // blank level representing global coarse problem to make further indexing easier

        int maxLevel = -1;
        for (size_t i=0; i<activeLevels_.size(); i++) {
            if (activeLevels_[i][0]) {
                maxLevel = i;
                break;
            }
        }

        const int coarsestLevel = maxLevel;

        for (size_t i=maxLevel+1; i<activeLevels_.size(); i++) {
            if (activeLevels_[i][0]) {
                // init local patch preconditioners on each level
                const auto& fineNetwork = *contactNetwork_.level(i);
                levelPatchPreconditioners_.push_back(std::make_shared<LevelPatchPreconditioner>(*contactNetwork_.level(maxLevel), fineNetwork, mode_));
                levelPatchPreconditioners_.back()->setPatchDepth(parset.get<size_t>("patchDepth"));
                maxLevel = i;
            }
        }

        levelMat_.resize(size()-1);
        levelX_.resize(size());
        levelRhs_.resize(size());

        // init patch solvers
        levelSolvers_.resize(size());
        levelItSteps_.resize(size());
        levelErrorNorms_.resize(size());
        //auto coarseStep =
        levelItSteps_[0] = Dune::Solvers::BlockGSStepFactory<MatrixType, VectorType>::createPtr(Dune::Solvers::BlockGS::LocalSolvers::direct(0.0));
        levelErrorNorms_[0] = std::make_shared<EnergyNorm<MatrixType, VectorType>>(*levelItSteps_[0].get());
        levelSolvers_[0] = /*std::make_shared<PatchSolver>(*levelItSteps_[0].get(),
                                   parset.get<size_t>("maximumIterations"),
                                   parset.get<double>("tolerance"),
                                   *levelErrorNorms_[0].get(),
                                   parset.get<Solver::VerbosityMode>("verbosity"),
                                   false); // absolute error */
                std::make_shared<PatchSolver>();

        for (size_t i=1; i<levelSolvers_.size(); i++) {
            //auto gsStep = Dune::Solvers::BlockGSStepFactory<MatrixType, VectorType>::create(Dune::Solvers::BlockGS::LocalSolvers::direct(0.0));
            levelItSteps_[i] = Dune::Solvers::BlockGSStepFactory<MatrixType, VectorType>::createPtr(Dune::Solvers::BlockGS::LocalSolvers::direct(0.0));
            levelErrorNorms_[i] = std::make_shared<EnergyNorm<MatrixType, VectorType>>(*levelItSteps_[i].get());
            levelSolvers_[i] = /*std::make_shared<PatchSolver>(*levelItSteps_[i].get(),
                                       parset.get<size_t>("maximumIterations"),
                                       parset.get<double>("tolerance"),
                                       *levelErrorNorms_[i].get(),
                                       parset.get<Solver::VerbosityMode>("verbosity"),
                                       false); // absolute error*/
                    std::make_shared<PatchSolver>();
            levelPatchPreconditioners_[i]->setPatchSolver(levelSolvers_[i]);
        }


        // init multigrid transfer
        mgTransfer_.resize(levelPatchPreconditioners_.size()-1);
        for (size_t i=0; i<mgTransfer_.size(); i++) {
            mgTransfer_[i] = std::make_shared<MGTransfer>();
            mgTransfer_[i]->setup(contactNetwork_, levelPatchPreconditioners_[i+1]->level(), levelPatchPreconditioners_[i+1]->fineLevel());
        }

        // Remove any recompute filed so that initially the full transferoperator is assembled
        for (size_t i=0; i<mgTransfer_.size(); i++)
            std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->setRecomputeBitField(nullptr);

        // specify the subset of entries to be reassembled at each iteration
       /* recompute_.resize(size());
        recompute_[recompute_.size()-1] = contactNetwork_.nBodyAssembler().totalHasObstacle_;
        for (int i=mgTransfer_.size()-1; i>=0; i--) {
            std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->restrict(recompute_[i+1], recompute_[i]);
          //  std::dynamic_pointer_cast<DenseMultigridTransfer<VectorType, BitVector, MatrixType> >(mgTransfer_[i])->restrict(recompute_[i+1], recompute_[i]);
            std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->setRecomputeBitField(&recompute_[i]);
            print(recompute_[i], "recompute: ");
        }*/
    }

   void setMatrix(const MatrixType& mat) override {
       Base::setMatrix(mat);

       ignoreHierarchy_.resize(size());
       ignoreHierarchy_.back() = Base::ignore();
       for (int i=mgTransfer_.size()-1; i>=0; i--) {
           std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->restrictToFathers(ignoreHierarchy_[i+1], ignoreHierarchy_[i]);
       }

      /* truncation_.resize(size());
       truncation_[size()-1] = Base::ignore();
       print(truncation_.back(), "truncation: ");
       for (int i=mgTransfer_.size()-1; i>=0; i--) {
           std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->restrict(truncation_[i+1], truncation_[i]);
           std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->setCriticalBitField(&truncation_[i+1]);
           print(truncation_[i], "truncation: ");
       }*/

        size_t maxLevel = levelPatchPreconditioners_.size()-1;
        levelPatchPreconditioners_[maxLevel]->setMatrix(mat);

        //print(mat, "levelMat_" + std::to_string(maxLevel) + ":");

        for (int i=levelPatchPreconditioners_.size()-2; i>0; i--) {
            const auto& transfer = *mgTransfer_[i];

            //print(*levelPatchPreconditioners_[i+1]->getMatrix(), "levelMat:");
            //print(levelMat_[i], "coarseMat:");

            transfer.galerkinRestrictSetOccupation(*levelPatchPreconditioners_[i+1]->getMatrix(), levelMat_[i]);
            transfer.galerkinRestrict(*levelPatchPreconditioners_[i+1]->getMatrix(), levelMat_[i]);
            levelPatchPreconditioners_[i]->setMatrix(levelMat_[i]);

            //print(levelMat_[i], "levelMat_" + std::to_string(i) + ":");

            // Set solution vector sizes for the lower levels
            Dune::MatrixVector::resize(levelX_[i], levelMat_[i]);
            //Dune::MatrixVector::resize(levelRhs_[i], levelMat_[i]);
        }

        // set coarse global problem
        const auto& coarseTransfer = *mgTransfer_[0];
        const auto& fineMat = *levelPatchPreconditioners_[1]->getMatrix();
        coarseTransfer.galerkinRestrictSetOccupation(fineMat, levelMat_[0]);
        coarseTransfer.galerkinRestrict(fineMat, levelMat_[0]);

        //print(levelMat_[0], "levelMat_0:");

        /*dynamic_cast<Base&>(levelSolvers_[0]->getIterationStep()).setMatrix(levelMat_[0]); */

        Dune::MatrixVector::resize(levelX_[0], levelMat_[0]);
         //Dune::MatrixVector::resize(levelRhs_[0], levelMat_[0]);
   }

   void setTruncation(const BitVector& truncation) {
       std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_.back())->setCriticalBitField(&truncation);
   }

   /*void setProblem(const MatrixType& mat, VectorType& x, const VectorType& rhs) override {
       LinearIterationStep<MatrixType, VectorType>::setProblem(mat, x, rhs);

       truncation_.resize(size());
       truncation_[size()-1] = this->ignore();
       print(truncation_.back(), "truncation: ");
       for (int i=mgTransfer_.size()-1; i>=0; i--) {
           std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->restrict(truncation_[i+1], truncation_[i]);
           std::dynamic_pointer_cast<TruncatedMGTransfer<VectorType> >(mgTransfer_[i])->setCriticalBitField(&truncation_[i]);
           print(truncation_[i], "truncation: ");
       }

        size_t maxLevel = levelPatchPreconditioners_.size()-1;
        levelX_[maxLevel] = x;
        levelRhs_[maxLevel] = rhs;
        levelPatchPreconditioners_[maxLevel]->setProblem(mat, levelX_[maxLevel], levelRhs_[maxLevel]);

        for (int i=levelPatchPreconditioners_.size()-2; i>0; i--) {
            const auto& transfer = *mgTransfer_[i];

            transfer.restrict(levelX_[i+1], levelX_[i]);
            transfer.restrict(levelRhs_[i+1], levelRhs_[i]);
            transfer.galerkinRestrict(*levelPatchPreconditioners_[i+1]->getMatrix(), levelMat_[i]);

            levelPatchPreconditioners_[i]->setProblem(levelMat_[i], levelX_[i], levelRhs_[i]);
        }

        // set coarse global problem
        const auto& coarseTransfer = *mgTransfer_[0];

        coarseTransfer.restrict(levelX_[1], levelX_[0]);
        coarseTransfer.restrict(levelRhs_[1], levelRhs_[0]);
        coarseTransfer.galerkinRestrict(levelMat_[1], levelMat_[0]);

        dynamic_cast<LinearIterationStep<MatrixType, VectorType>&>(levelSolvers_[0]->getIterationStep()).setProblem(levelMat_[0], levelX_[0], levelRhs_[0]);

      //  EnergyNorm<MatrixType, VectorType> coarseErrorNorm(levelSolvers_[0]->getIterationStep());
      //  levelSolvers_[0]->setErrorNorm(coarseErrorNorm);
    } */

    void iterate() {
        size_t maxLevel = levelPatchPreconditioners_.size()-1;
        levelX_[maxLevel] = *this->getIterate();
        levelRhs_[maxLevel] = *Base::rhs_;

        for (int i=levelPatchPreconditioners_.size()-2; i>0; i--) {
            const auto& transfer = *mgTransfer_[i];

            /*print(levelX_[i+1], "levelX_:");
            print(levelX_[i], "levelX_:");
            print(levelRhs_[i+1], "levelRhs_:");
            print(levelRhs_[i], "levelRhs_:");*/

            transfer.restrict(levelX_[i+1], levelX_[i]);
            transfer.restrict(levelRhs_[i+1], levelRhs_[i]);
        }

        // set coarse global problem
        const auto& coarseTransfer = *mgTransfer_[0];
        coarseTransfer.restrict(levelX_[1], levelX_[0]);
        coarseTransfer.restrict(levelRhs_[1], levelRhs_[0]);


        if (mode_ == additive)
            iterateAdd();
        else
            iterateMult();
    }

    void iterateAdd() {
        //print(this->ignore(), "ignore: ");

        *(this->x_) = 0;	

        VectorType x;

        /*
        // solve coarse global problem
        auto& coarseSolver = levelSolvers_[0];
        //coarseSolver->getIterationStep().setIgnore(ignoreHierarchy_[0]);

        dynamic_cast<LinearIterationStep<MatrixType, VectorType>&>(levelSolvers_[0]->getIterationStep()).setProblem(levelMat_[0], levelX_[0], levelRhs_[0]);

        coarseSolver->check();
        coarseSolver->preprocess();
        coarseSolver->solve();

        mgTransfer_[0]->prolong(coarseSolver->getIterationStep().getSol(), x); */


        BitVector emptyIgnore0(levelX_[0].size(), false);
        LocalProblem<MatrixType, VectorType> localProblem(levelMat_[0], levelRhs_[0], ignoreHierarchy_[0]);
        Vector newR;
        localProblem.getLocalRhs(levelX_[0], newR);

        auto& coarseSolver = levelSolvers_[0];
        coarseSolver->setProblem(localProblem.getMat(), levelX_[0], newR);
        coarseSolver->preprocess();
        coarseSolver->solve();

        mgTransfer_[0]->prolong(levelX_[0], x);

        //print(levelX_[0], "MultilevelPatchPreconditioner: levelX_[0]");
        //print(x, "MultilevelPatchPreconditioner: x0");


        for (size_t i=1; i<levelPatchPreconditioners_.size()-1; i++) {
            const auto& transfer = *mgTransfer_[i];
            auto& preconditioner = *levelPatchPreconditioners_[i];

            preconditioner.setIgnore(ignoreHierarchy_[i]);
            preconditioner.apply(levelX_[i], levelRhs_[i]);

            //print(levelX_[i], "MultilevelPatchPreconditioner: levelX_[i] " + std::to_string(i));

            x += preconditioner.getSol();

            VectorType newX;
            transfer.prolong(x, newX);

            //print(newX, "MultilevelPatchPreconditioner: newX " + std::to_string(i));

            x = newX;

            //print(x, "MultilevelPatchPreconditioner: x " + std::to_string(i));
        }

        auto& preconditioner = *levelPatchPreconditioners_.back();

        preconditioner.setIgnore(ignoreHierarchy_.back());
        preconditioner.apply(levelX_.back(), levelRhs_.back());

        x += preconditioner.getSol();

        //print(x, "MultilevelPatchPreconditioner: xmax");


       /* BitVector emptyIgnore(this->x_->size(), false);
        LocalProblem<MatrixType, VectorType> globalProblem(*this->mat_.get(), *this->rhs_, emptyIgnore);
        globalProblem.getLocalRhs(*this->x_, newR);
        VectorType totalX = *(this->x_);
        auto& globalSolver = levelSolvers_.back();
        globalSolver->setProblem(globalProblem.getMat(), totalX, newR);
        globalSolver->preprocess();
        globalSolver->solve();

        print(totalX, "MultilevelPatchPreconditioner: totalX"); */

        *(this->x_) = x;

        //print(*(this->x_), "MultilevelPatchPreconditioner: sol");
    }


    void iterateMult() {
        *(this->x_) = 0;

        /*
        VectorType x;

        for (int i=(levelPatchPreconditioners_.size()-1); i>=0; i--) {
            VectorType updatedRhs(*(this->rhs_));
            this->mat_->mmv(*(this->x_), updatedRhs);

            mgTransfer_[i]->restrict(*(this->x_), levelX_[i]);
            mgTransfer_[i]->restrict(updatedRhs, levelRhs_[i]);

            levelPatchPreconditioners_[i]->setProblem(*(this->mat_), levelX_[i], levelRhs_[i]);

            levelPatchPreconditioners_[i]->iterate();
            const VectorType& it = levelPatchPreconditioners_[i]->getSol();

            mgTransfer_[i]->prolong(it, x);

            *(this->x_) += x;
        }

        VectorType updatedCoarseRhs(*(this->rhs_));
        this->mat_->mmv(*(this->x_), updatedCoarseRhs);

        size_t j = levelPatchPreconditioners_.size();
        mgTransfer_[j]->restrict(*(this->x_), levelX_[j]);
        mgTransfer_[j]->restrict(updatedCoarseRhs, levelRhs_[j]);

        coarseGlobalPreconditioner_->setProblem(*(this->mat_), levelX_[j], levelRhs_[j]);
        coarseGlobalPreconditioner_->iterate();
        const VectorType& it = coarseGlobalPreconditioner_->getSol();

        mgTransfer_[j]->prolong(it, x);

        *(this->x_) += x; */
    }

    void build() {
        for (size_t i=1; i<levelPatchPreconditioners_.size(); i++) {
            levelPatchPreconditioners_[i]->build();
        }
    }

    void setPatchDepth(const size_t patchDepth = 0) {
        for (size_t i=1; i<levelPatchPreconditioners_.size(); i++) {
            levelPatchPreconditioners_[i]->setPatchDepth(patchDepth);
        }
    }

    auto contactNetwork() const {
        return contactNetwork_;
    }

    size_t size() const {
        return levelPatchPreconditioners_.size();
    }
};

#endif

